// PrestoServiceHost.cs
//
// PRESTO Starter Kit for Microsoft .NET 3.0
// 2007 Microsoft France
// Published under the CECILL-B Free Software license agreement.
// (http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.txt)
// 
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, 
// WHETHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED 
// WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE. 
// THE ENTIRE RISK OF USE OR RESULTS IN CONNECTION WITH THE USE OF THIS CODE 
// AND INFORMATION REMAINS WITH THE USER. 
//

using System;
using System.Collections.Generic;
using System.Text;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Dispatcher;
using System.ServiceModel.Description;
using System.Diagnostics;

using System.ServiceModel.Security;
using System.Security.Cryptography.X509Certificates;
using System.Net.Security;

using dgme.finances.gouv.fr.presto;

using Microsoft.Presto.StarterKit.Constants;

namespace Microsoft.Presto.StarterKit
{
    public class PrestoServiceHost
    {
        private ServiceHost serviceHost;
        
        private PrestoSoapNodeConfig config;
        public PrestoSoapNodeConfig Config
        {
            get { return config; }
            set { config = value; }
        }

        public bool IsListening
        {
            get { return (bool)(serviceHost != null); }
        }
  
        public Binding Binding
        {
            get { return config.GenerateBinding(); }
        }

        public PrestoServiceHost()
        {
            serviceHost = null;
        }

        void serviceHost_UnknownMessageReceived(object sender, UnknownMessageReceivedEventArgs e)
        {
            //throw new Exception("The method or operation is not implemented.");
            Console.WriteLine("serviceHost_UnknownMessageReceived");
            Console.WriteLine(e.Message.ToString());
        }

        internal void StartPrestoInterface()
        {
            if (serviceHost != null) StopPrestoInterface();

            Binding usedBinding = null;

            if (config.UseConfigurationFile)
            {
                this.config.AddressingVersion = PrestoAddressingVersion.WSAddressingAugust2004;
                this.config.UseDigitalSignature = false;
                this.config.UseEncryption = false;
                    
                usedBinding = new CustomBinding(config.GetBindingNameFromConfigurationFile()); 
            }
            else
            {
                usedBinding = Binding;
            }
            Uri baseAddress = null;
            try
            {
                // Create a service host
                baseAddress = new Uri(config.ServiceEndpointUri);
                // Following modification is required for use in conjunction with TCP Trace
                //serviceHost = new ServiceHost(typeof(PrestoService), baseAddress);
                serviceHost = new ServiceHost(typeof(PrestoService), new Uri(config.ServiceEndpointMexUri));
            }
            catch (Exception ex)
            {
                Trace.WriteLine(ex.Message);
                if (ex.InnerException != null) Trace.WriteLine(ex.InnerException.Message);
            }

            // Add a MEX endpoint for the service
            ServiceMetadataBehavior metadataBehavior = new ServiceMetadataBehavior();
            metadataBehavior.HttpGetEnabled = true;
            metadataBehavior.HttpGetUrl = new Uri(config.ServiceEndpointMexUri);
            serviceHost.Description.Behaviors.Add(metadataBehavior);
            serviceHost.AddServiceEndpoint(typeof(IMetadataExchange), MetadataExchangeBindings.CreateMexHttpBinding(), 
                                           config.ServiceEndpointMexUri);

            // Add an endpoint for the IPresto Interface (Reliable One-Way MEP) using the AddEndpoint helper method
            // to create the ServiceEndpoint and add it to the ServiceDescription
            ServiceEndpoint serviceEndpoint;
            if (config.UseSoapIntermediary)
            {
                // Following modification is required for use in conjunction with TCP Trace
                //serviceEndpoint = serviceHost.AddServiceEndpoint(typeof(IPresto), usedBinding, EndpointUri.oneWayRelativeAddress, new Uri(config.SoapIntermediaryEndpointUri));
                serviceEndpoint = serviceHost.AddServiceEndpoint(typeof(IPresto), usedBinding, EndpointUri.nullAddress);
                serviceEndpoint.ListenUri = new Uri(config.SoapIntermediaryEndpointUri);
                //

                // When used in conjunction with WS-RM, this option fails (no way to handle trace with TCPTrace,
                // because an Offer tag is required by the RM destination (there should not be any, because the contract 
                // scope is OneWay). Without the use of TCPTrace, the service does not ask for an offer tag.

                // An error has been encountered: Type : System.ServiceModel.ProtocolException
                // The request to create a reliable session has been refused by the RM Destination. 
                // The endpoint at http://localhost:9090/ processes reply sessions. 
                // The create sequence request must contain an offer for a return sequence. 
                // This is likely caused by a binding mismatch. The channel could not be opened


                serviceEndpoint.Address = new EndpointAddress(new Uri(config.ServiceEndpointUri + "/" + EndpointUri.oneWayRelativeAddress));
            }
            else
            {
                // Following modification is required for use in conjunction with TCP Trace (serviceHost is no longer created with baseAdress)
                //serviceEndpoint = serviceHost.AddServiceEndpoint(typeof(IPresto), usedBinding, EndpointUri.oneWayRelativeAddress);
                serviceEndpoint = serviceHost.AddServiceEndpoint(typeof(IPresto), usedBinding, EndpointUri.nullAddress);
                serviceEndpoint.Address = new EndpointAddress(new Uri(config.ServiceEndpointUri + "/" + EndpointUri.oneWayRelativeAddress));
            }

            // Add an endpoint for the IPresto2 Interface (Reliable Request-Reply MEP)
            ServiceEndpoint serviceEndpoint2;
            if (config.UseSoapIntermediary) 
            {
                // Following modification is required for use in conjunction with TCP Trace
                //serviceEndpoint2 = serviceHost.AddServiceEndpoint(typeof(IPresto2), usedBinding, EndpointUri.requestReplyRelativeAddress, new Uri(config.SoapIntermediaryEndpointUri)); 
                serviceEndpoint2 = serviceHost.AddServiceEndpoint(typeof(IPresto2), usedBinding, EndpointUri.nullAddress);
                serviceEndpoint2.ListenUri = new Uri(config.SoapIntermediaryEndpointUri);
                serviceEndpoint2.Address = new EndpointAddress(new Uri(config.ServiceEndpointUri + "/" + EndpointUri.requestReplyRelativeAddress));
            }
            else
            {
                // Following modification is required for use in conjunction with TCP Trace (serviceHost is no longer created with baseAdress)
                //serviceEndpoint2 = serviceHost.AddServiceEndpoint(typeof(IPresto2), usedBinding, EndpointUri.requestReplyRelativeAddress);
                serviceEndpoint2 = serviceHost.AddServiceEndpoint(typeof(IPresto2), usedBinding, EndpointUri.nullAddress);
                serviceEndpoint2.Address = new EndpointAddress(new Uri(config.ServiceEndpointUri + "/" + EndpointUri.requestReplyRelativeAddress));
            }


            if (config.UseReliableMessaging)
            {
                ContractDescription contractDescription = serviceEndpoint.Contract;
                ContractDescription contractDescription2 = serviceEndpoint2.Contract;

                contractDescription.SessionMode = SessionMode.Required;
                contractDescription2.SessionMode = SessionMode.Required;

                OperationDescription operationDescription = contractDescription.Operations[0];
                OperationDescription operationDescription2 = contractDescription2.Operations[0];

                operationDescription.IsInitiating = true;
                operationDescription2.IsInitiating = true;
                operationDescription.IsTerminating = true;
                operationDescription2.IsTerminating = true;
            }

            // Support of several authentication modes
            if (!(config.NoAuthentication))
            {
                // Extract the ServiceCredentials behaviour or create one
                ServiceCredentials serviceCredentials = this.serviceHost.Description.Behaviors.Find<ServiceCredentials>();
                if (serviceCredentials == null)
                {
                    serviceCredentials = new ServiceCredentials();
                    this.serviceHost.Description.Behaviors.Add(serviceCredentials);
                }

                // Set the Service X.509 certificate
                serviceCredentials.ServiceCertificate.SetCertificate(StoreLocation.LocalMachine, StoreName.My, 
                    X509FindType.FindByThumbprint, config.ServiceCertificateThumbprint);

                if (config.MutualCertificate)
                {
                    // MutualCertificate Authentication Scenario
                    // Setting the CertificateValidationMode to PeerTrust or PeerOrChainTrust means that if the  
                    // certificate is in the Trusted People store, then it will be trusted without performing a validation 
                    // of the certificate's issuer chain. Such a setting is used here for convenience so that the 
                    // sample can be run without having to have certificates issued by a certificate authority (CA).
                    // This setting is less secure than the default, ChainTrust. The security implications of this 
                    // setting should be carefully considered before using PeerTrust or PeerOrChainTrust in production code. 
                    serviceCredentials.ClientCertificate.Authentication.CertificateValidationMode =
                        X509CertificateValidationMode.ChainTrust;

                    // The PRESTO Starter Kit X.509 Certificates have been generated with the Certificate Creation tool
                    // (makecert.exe). These certificates are for development/testing purposes only and do NOT have a CDP
                    // extension. As a result, the revocation test has to been turned off, otherwise, such certificates are
                    // considered expired. In production code, the following line MUST be commented.
                    serviceCredentials.ClientCertificate.Authentication.RevocationMode = X509RevocationMode.NoCheck;
                }

                ContractDescription contractDescription = serviceEndpoint.Contract;
                ContractDescription contractDescription2 = serviceEndpoint2.Contract;
                // Support of XMLENC is currently not part of the PRESTO specification version 1.0
                if (config.UseEncryption)
                {
                    contractDescription.ProtectionLevel = ProtectionLevel.EncryptAndSign;
                    contractDescription2.ProtectionLevel = ProtectionLevel.EncryptAndSign;
                }
                else if (config.UseDigitalSignature)
                {
                    contractDescription.ProtectionLevel = ProtectionLevel.Sign;
                    contractDescription2.ProtectionLevel = ProtectionLevel.Sign;
                }
                else 
                {
                    contractDescription.ProtectionLevel = ProtectionLevel.None;
                    contractDescription2.ProtectionLevel = ProtectionLevel.None;
                }
            }

            serviceHost.UnknownMessageReceived += new EventHandler<UnknownMessageReceivedEventArgs>(serviceHost_UnknownMessageReceived);
            
            // Create and open the service runtime
            serviceHost.Open();
            DisplayServiceEndpointsState();
        }

        public void StopPrestoInterface()
        {
            if (serviceHost != null)
            {
                serviceHost.Close();
                DisplayServiceEndpointsState();
                serviceHost = null;
            }
        }

        private void DisplayServiceEndpointsState()
        {
            // Display the Service endpoints state
            OutputTrace.Log("Presto Service Endpoints:"); 
            foreach (ChannelDispatcher channelDispatcher in serviceHost.ChannelDispatchers)
            {
                foreach (EndpointDispatcher endpointDispatcher in channelDispatcher.Endpoints)
                {
                    StringBuilder sb = new StringBuilder();
                    sb.AppendFormat("Listener ({0}, {1}) = {2}, State = {3}",
                            endpointDispatcher.ContractName.ToString(),
                            endpointDispatcher.ContractNamespace.ToString(), 
                            endpointDispatcher.EndpointAddress.ToString(), 
                            channelDispatcher.State.ToString());
                    OutputTrace.Log(sb.ToString());
                }
            }
            OutputTrace.Log(Environment.NewLine); 
        }
    }
}

